"""Settings Module."""

from scphylo.settings._settings import (
    hg19,
    hg38,
    logfile,
    m10x,
    mm10,
    refs,
    tools,
    verbosity,
)

__all__ = (
    hg19,
    hg38,
    logfile,
    m10x,
    mm10,
    refs,
    tools,
    verbosity,
)
