"""Datasets Module."""

from scphylo.datasets._datasets import (
    acute_lymphocytic_leukemia1,
    acute_lymphocytic_leukemia2,
    acute_lymphocytic_leukemia3,
    acute_lymphocytic_leukemia4,
    acute_lymphocytic_leukemia5,
    acute_lymphocytic_leukemia6,
    acute_myeloid_leukemia1,
    acute_myeloid_leukemia2,
    colorectal1,
    colorectal2,
    colorectal3,
    erbc,
    example,
    high_grade_serous_ovarian_cancer1,
    high_grade_serous_ovarian_cancer2,
    high_grade_serous_ovarian_cancer3,
    high_grade_serous_ovarian_cancer_3celllines,
    melanoma20,
    muscle_invasive_bladder,
    myeloproliferative_neoplasms18,
    myeloproliferative_neoplasms78,
    myeloproliferative_neoplasms712,
    oligodendroglioma_idh_mutated_tumor,
    renal_cell_carcinoma,
    test,
    tnbc,
)
from scphylo.datasets._simulate import add_doublets, add_noise, add_readcount, simulate

__all__ = (
    acute_lymphocytic_leukemia1,
    acute_lymphocytic_leukemia2,
    acute_lymphocytic_leukemia3,
    acute_lymphocytic_leukemia4,
    acute_lymphocytic_leukemia5,
    acute_lymphocytic_leukemia6,
    acute_myeloid_leukemia1,
    acute_myeloid_leukemia2,
    colorectal1,
    colorectal2,
    colorectal3,
    erbc,
    example,
    high_grade_serous_ovarian_cancer1,
    high_grade_serous_ovarian_cancer2,
    high_grade_serous_ovarian_cancer3,
    high_grade_serous_ovarian_cancer_3celllines,
    melanoma20,
    muscle_invasive_bladder,
    myeloproliferative_neoplasms18,
    myeloproliferative_neoplasms78,
    myeloproliferative_neoplasms712,
    oligodendroglioma_idh_mutated_tumor,
    renal_cell_carcinoma,
    test,
    tnbc,
    simulate,
    add_noise,
    add_doublets,
    add_readcount,
)
