"""scPhylo Module."""

from scphylo import datasets, io
from scphylo import logging as logg
from scphylo import pl, pp, settings, tl, ul

__author__ = ", ".join(["Farid Rashidi"])
__maintainer__ = ", ".join(["Farid Rashidi"])
__email__ = ", ".join(["farid.rsh@gmail.com"])
__version__ = "0.0.5"
__all__ = (datasets, io, logg, pl, pp, settings, tl, ul)
